// *********************************************************************
//
//	unzip.c	-	main source for Windows unzipper  (TABS == 4)
//
// *********************************************************************

#include "unzip.h"
#include <ctype.h>
#include <errno.h>

BOOL read_zip_files( void );
BOOL extract_this_file( char *fn );
void unzip( void );
void do_unzip( void );
// int unzip_file(PSZ pszZipFile, PSZ pszDestinationDir = NULL);

// -----------------------------------------------
//	STRUCTS
// -----------------------------------------------


//
//	The union 'work' was transformed into a FAR memory object, in order
//	to allow for increases in the 'inbuf' and 'outbuf' arrays (which are
//	dynamically allocated).  All of the memory copy routines which
//	interfaced to the 'work area' were changed to _fmemcpy & _fmemset.
//
//
HGLOBAL		work_area						= NULL;
union 		work 							FAR *area;

struct		local_file_header				lrec;
struct		central_directory_file_header	crec;
struct		end_central_dir_record			ecrec;

// -----------------------------------------------
//	FLAGS
// -----------------------------------------------
//
//	This version understands only 3 different flags:
//
//	all_flag	This flag tells xzip to extract ALL files
//				(-a option on the command line)
//
//	rep_flag	This flag tells xzip to replace all dest. files
//				(-r option on the command line)
//
//	date_flag	this flag tells xzip to change the dates of
//				unzipped files to their archived date.
//				(-t option on the command line)
//
//	dir_flag	this flag tells xzip to create subdirectories
//				(-d option on the command line)
//

BOOL	all_flag	=	FALSE;
BOOL	rep_flag	=	TRUE;			// Default: replace existing files
BOOL	date_flag	=	FALSE;
BOOL	dir_flag	=	TRUE;			// Default: create directories
BOOL	ask_path_flag = TRUE;


// -----------------------------------------------
//	FILE stuff
// -----------------------------------------------

BOOL		overwrite_files	=	TRUE;

long		csize;
long		ucsize;

BOOL		global_error = FALSE;
char		global_filename[ PATH_LEN ];

// -----------------------------------------------
//	INPUT VARIABLES
// -----------------------------------------------
int			zip_in_file;			// Global handle for ZIP archive

int			incnt;
BYTE		*inbuf	=	NULL;			/* input buffer (any size is legal) and pointer */
BYTE		*inptr	=	NULL;

long		cur_zipfile_bufstart;   /* extract_or_test_files, readbuf, ReadByte */

ULONG		bitbuf;
int			bits_left;
BOOL		zipeof;

// -----------------------------------------------
//	OUTPUT VARIABLES
// -----------------------------------------------
int			zip_out_file;			// Global handle for current output file

BYTE		*outbuf		= NULL;					/* buffer for rle look-back */
BYTE		*outptr		= NULL;
BYTE		*outout		= NULL;					/* scratch pad for ASCII-native trans */
long		outpos;						/* absolute position in outfile */
int			outcnt;						/* current position in outbuf */

BOOL		mem_mode = FALSE;
BOOL		disk_full = FALSE;


UWORD mask_bits[] = {
    0x0000,
    0x0001, 0x0003, 0x0007, 0x000f, 0x001f, 0x003f, 0x007f, 0x00ff,
    0x01ff, 0x03ff, 0x07ff, 0x0fff, 0x1fff, 0x3fff, 0x7fff, 0xffff
};


// -----------------------------------------------
//	FILE NAMES
// -----------------------------------------------
char		source_directory[ PATH_LEN ];
char		destination_directory[ PATH_LEN ];

char		command_line[ PATH_LEN ];
char		zip_filename[ PATH_LEN ];
char		zip_out_filename[ PATH_LEN ];
char		option1[ 10 ];
char		option2[ 10 ];
char		option3[ 10 ];
char		option4[ 10 ];

HGLOBAL		hExtract_files;
int			extract_file_count;
int			total_file_count;
BOOL		extract_all = FALSE;

// *********************************************************************
//
//	VARS	(not related to unzipping - GUI stuff here)
//
// *********************************************************************
HANDLE		hInst;
HWND     cwnd = HWND_DESKTOP;
HWND		list_box;
HWND		message_box;
int			wwide, whigh;
int			font_width, font_height;
BOOL		commence;

char		err_msg[ 256 ];

//            unzip_file("c:\\winzip\\test.zip", "c:\\winzip\\update");

// *********************************************************************
//
//	out_msg()
//
// *********************************************************************
void out_msg( char *message )
{
	SetWindowText( message_box, message );
}

int unzip_file(PSZ pszZipFile, PSZ pszDestinationDir)
{
   long l;
   int a;
   DLGPROC dlgproc;

   memset(option1, 0, sizeof(option1));
   memset(option2, 0, sizeof(option2));
   memset(option3, 0, sizeof(option3));
   memset(option4, 0, sizeof(option4));
   memset(zip_filename, 0, sizeof(zip_filename));

	strcpy(zip_filename, pszZipFile);
   if (pszDestinationDir)
      {
      strcpy(destination_directory, pszDestinationDir);

      // Let's go ahead and make the destination dir if it doesn't exist.
		if (mkdir(destination_directory) && (EACCES != errno))
         {
         ::MessageBox( cwnd, "Could not create destination dir", "Error", MB_OK );
         }
      }
   else
      getcwd( destination_directory, PATH_LEN );

   if( destination_directory[ strlen(destination_directory)-1] == '\\' )
   {
      destination_directory[ strlen(destination_directory)-1] = 0;
   }
   strlwr( destination_directory );

   l = (long) (((long)HSIZE+2L)*4L + 16L);
   work_area = GlobalAlloc( GHND, l );
   if(  work_area == NULL )
   {
      ::MessageBox( cwnd, "Out of memory", "Error", MB_OK );
      return 0;
   }
   area = (union work FAR *)GlobalLock( work_area );

   inbuf = (unsigned char *) calloc( INBUFSIZ + 1, 1 );
   outbuf = (unsigned char *) calloc( OUTBUFSIZ + 1, 1 );

   if( inbuf == 0 || outbuf == 0 )
      {
      ::MessageBox( cwnd, "Out of memory: inbuf or outbuf", "Error", MB_OK );
      return 0;
      }
   outptr = outbuf;
   inptr = inbuf;


// if( ask_path_flag )
//    {
//    dlgproc = (DLGPROC) MakeProcInstance( (FARPROC)unzip_to_proc, hInst );
//    a = DialogBox( hInst, "DIALOG_1", cwnd, dlgproc );
//    FreeProcInstance( (FARPROC) dlgproc );
//    }
// else
//    {
      a = 1;
//    }

   if( a && read_zip_files() )
   {
      unzip();
      out_msg( "Done!" );
   }

	if( work_area )
   {
		GlobalUnlock( work_area );
        GlobalFree( work_area );
	}

	if( hExtract_files )
		GlobalFree( hExtract_files );

	if( inbuf )
		free( inbuf );
	if( outbuf )
    	free( outbuf );

   return(0);
}

// *********************************************************************
//
//	read_zip_files	- read the files to extract from the INI file
//
// *********************************************************************
BOOL read_zip_files( void )
{
int			fp1;
struct		ex_file	FAR *extract_files;
int			a, count = 0;
char		buf[ PATH_LEN ];
char		tbuf[ 10 ];
BOOL		res	=	TRUE;
long		flen;
long		offset;

	// -----------------------------------------------
	//	Read files to extract from INI file
	// -----------------------------------------------
	count = GetPrivateProfileInt( "files", "count", 0, "xzip.ini" );
//	count = 0;
	if( count > 0 )
    {
		if( hExtract_files = GlobalAlloc( GHND, sizeof( struct ex_file ) * count ) )
		{
			extract_files = (struct ex_file FAR *)GlobalLock( hExtract_files );
			for( a=0; a<count; a++ )
			{
				sprintf( tbuf, "%d", a+1 );
				GetPrivateProfileString( "Files", tbuf, "", buf, PATH_LEN, "xzip.ini" );
				_fstrcpy( extract_files[a].file_name, (char FAR *)buf );
		    }   	
			GlobalUnlock( hExtract_files );
			extract_file_count	= count;
			extract_all = FALSE;
		}else{
			extract_file_count	= 0;
			extract_all = TRUE;
         ::MessageBox( cwnd, "Memory error! Can't allocate ex_file buffer", "Error", MB_OK );
		}
	}else{
		extract_all = TRUE;
	}


	// -----------------------------------------------
	//	count # files in zip archive
	// -----------------------------------------------
	total_file_count = 0;
	if( (fp1 = open( zip_filename, O_RDONLY | O_BINARY ))>0 )
	{
		flen = filelength( fp1 );
		a = 1;
		while( a )
		{
			read( fp1, &lrec,  sizeof( struct local_file_header ) );
			flen -= (long)( sizeof( struct local_file_header ) );
			if( lrec.ID[0] == 'P' && lrec.ID[1] == 'K' && lrec.ID[2] == 3 && lrec.ID[3] == 4 )
			{
				total_file_count++;
				sprintf( err_msg, "File: %d", total_file_count );
				out_msg( err_msg );
				offset = (long)(lrec.compressed_size + lrec.filename_length + lrec.extra_field_length);
				lseek( fp1, offset, SEEK_CUR );
				flen -= offset;
			}else{
				a = 0;
				offset = (long)(lrec.compressed_size + lrec.filename_length + lrec.extra_field_length);
				lseek( fp1, offset, SEEK_CUR );
				flen -= offset;
			}
			if( flen <= (long)0 )
				a = 0;
		}
		close( fp1 );

//		sprintf( err_msg, "Total %d files", total_file_count );
//    ::MessageBox( cwnd, err_msg, "Status", MB_OK );

	}else{
		sprintf( err_msg, "Error opening ZIP file: %s", zip_filename );
      ::MessageBox( cwnd, err_msg, "Error", MB_OK );
		res = FALSE;
	}

	sprintf( err_msg, "Found %d ZIPPED files", total_file_count );
	out_msg( err_msg );

	return res;

}


// *********************************************************************
//
//	extract_this_file?
//
// *********************************************************************
BOOL extract_this_file( char *fn )
{
struct		ex_file	FAR *extract_files;
int			a;
char		buf[ PATH_LEN ];
char		*p;
BOOL		res = FALSE;

	if( fn[ strlen(fn)-1 ] == '\\' )
		return FALSE;


	//	Build the output path+filename.
	strcpy( zip_out_filename, destination_directory );
	strcat( zip_out_filename, "\\" );
	strcat( zip_out_filename, fn );

	if( all_flag )
    	return TRUE;

	//	Search the extract list for this filename
	if( !extract_all )
	{
		extract_files = (struct ex_file FAR *)GlobalLock( hExtract_files );
		for( a=0; a<extract_file_count; a++ )
		{
			_fstrcpy( (char FAR *)buf, extract_files[a].file_name );
			if( !stricmp( fn, buf ) )
			{
				res = TRUE;
				a = extract_file_count;
			}
		}
		GlobalUnlock( hExtract_files );
	}else{
		res = TRUE;
	}


	// File not found in extract list - return
	if( !res )
    	return res;

	//	If file exists, test for overwrite option
	if( overwrite_files )
		return res;

	return res;

}


// *********************************************************************
//
//	unzip - the unzip routine!
//
// *********************************************************************
void unzip( void )
{
int		a;
char	tbuf[ PATH_LEN ], *p;
long	pos1, pos2;

	if( (zip_in_file = open( zip_filename, O_RDONLY | O_BINARY )) == -1)
	{
		sprintf( err_msg, "Error opening ZIP file: %s", zip_filename );
      ::MessageBox( cwnd, err_msg, "Error", MB_OK );
		return;
	}

	for( a=0; a<total_file_count; a++ )
	{
		read( zip_in_file, &lrec, sizeof( struct local_file_header ) );
		if( lrec.ID[0] == 'P' && lrec.ID[1] == 'K' && lrec.ID[2] == 3 && lrec.ID[3] == 4 )
		{
			memset( tbuf, 0, PATH_LEN );
			read( zip_in_file, tbuf, lrec.filename_length );
			while( p = strchr( tbuf, '/' ) )
				*p = '\\';
			strcpy( global_filename, tbuf );

			if( lrec.extra_field_length )
				lseek( zip_in_file, (long)lrec.extra_field_length, SEEK_CUR );

			if( extract_this_file( tbuf ) )
			{
            	if( open_output_file() )
				{
					do_unzip();
					close( zip_out_file );
					if( global_error )
                    {
						sprintf( err_msg, "FATAL Error with file: %s", tbuf );
                  ::MessageBox( cwnd, err_msg, "Error", MB_OK );
						a = total_file_count;
					}
				}else{
					sprintf( err_msg, "Error creating file: %s", tbuf );
               ::MessageBox( cwnd, err_msg, "Error", MB_OK );
					lseek( zip_in_file, (long)lrec.compressed_size, SEEK_CUR );
				}
			}else{
				lseek( zip_in_file, (long)lrec.compressed_size, SEEK_CUR );
//				sprintf( err_msg, "Skipping:\t%s", tbuf );
//				out_msg( err_msg );
			}

		}else{
			a = total_file_count;
		}
	}
	close( zip_in_file );
}



// *********************************************************************
//
//	do_unzip - call the appropriate unzip routine to extract this file
//
// *********************************************************************
void do_unzip( void )
{
UWORD	b;
BYTE	mybyte;
int		a;

	csize		= (long) lrec.compressed_size;
	ucsize		= (long) lrec.uncompressed_size;

	disk_full	= FALSE;

	bits_left = 0;
    bitbuf = 0;

	switch (lrec.compression_method)
	{
		// -----------------------------------------------
		//	STORED
		// -----------------------------------------------
		case STORED:
			sprintf( err_msg, "Restoring:\t%s", zip_out_filename );
			out_msg( err_msg );

			a = min( lrec.compressed_size, (unsigned long) INBUFSIZ );
			while( (long)lrec.compressed_size > (long)0 )
			{
				read( zip_in_file, inbuf, a );
				write( zip_out_file, inbuf, a );
				lrec.compressed_size -= (long)a;
				a = min( lrec.compressed_size, (unsigned long) a );
			}
			break;

		// -----------------------------------------------
		//	SHRUNK
		// -----------------------------------------------
		case SHRUNK:
			sprintf( err_msg, "Unshrinking:\t%s", zip_out_filename );
			out_msg( err_msg );
			unShrink();
			FlushOutput();
	        break;

		// -----------------------------------------------
		//	REDUCED
		// -----------------------------------------------
		case REDUCED1:
    	case REDUCED2:
	    case REDUCED3:
    	case REDUCED4:
			sprintf( err_msg, "UnReducing:\t%s", zip_out_filename );
			out_msg( err_msg );
			unReduce();
			FlushOutput();
			break;

		// -----------------------------------------------
		//	IMPOLODED
		// -----------------------------------------------
		case IMPLODED:
			sprintf( err_msg, "Exploding:\t%s", zip_out_filename );
			out_msg( err_msg );
			explode();
			FlushOutput();
			break;


		// -----------------------------------------------
		//	DEFLATED
		// -----------------------------------------------
		case DEFLATED:
			sprintf( err_msg, "Inflating:\t\t%s", zip_out_filename );
			out_msg( err_msg );
			inflate();
			FlushOutput();
			break;

		// -----------------------------------------------
		//	UNKNOWN
		// -----------------------------------------------
		default:
        	sprintf( err_msg, "Unknown compression method: %d", lrec.compression_method );
         ::MessageBox( cwnd, err_msg, "Error", MB_OK );
			lseek( zip_in_file, lrec.compressed_size, SEEK_CUR );
			break;
	}

}
