// *********************************************************************
//
//	unzip.h
//
// *********************************************************************

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dir.h>
#include <direct.h>
#include <fcntl.h>
#include <io.h>
#include <sys\stat.h>

#include "os2type.h"

#define WM_COMMENCE (WM_USER+32)

#define UBYTE unsigned char
#define UWORD unsigned int
#define ULONG unsigned long

#define MAX_BITS      13                 /* used in unShrink() */
#define HSIZE         (1 << MAX_BITS)    /* size of global work area */
#define WSIZE 32768

// -----------------------------------------------
//	SIZES
// -----------------------------------------------
#define PATH_LEN		192
#define FILENAME_LEN	16

#define INBUFSIZ		16000		/* works for MS-DOS small model */
#define OUTBUFSIZ		INBUFSIZ


// -----------------------------------------------
//	COMPRESSION METHODS
// -----------------------------------------------
#define STORED            0
#define SHRUNK            1
#define REDUCED1          2
#define REDUCED2          3
#define REDUCED3          4
#define REDUCED4          5
#define IMPLODED          6
#define TOKENIZED         7
#define DEFLATED          8


// -----------------------------------------------
//	ex_file struct
// -----------------------------------------------
struct	ex_file {
	char	file_name[ PATH_LEN ];
	char	file_node[ FILENAME_LEN ];
};	


// *********************************************************************
//
//	PROTOS
//
// *********************************************************************
extern	void out_msg( char *message );


int unzip_file(PSZ pszZipFile, PSZ pszDestinationDir = NULL);
extern	void unShrink( void );
extern	void unReduce( void );
extern	int explode( void );
extern	void inflate( void );

// -----------------------------------------------
//	FILE IO PROTOS
// -----------------------------------------------
extern	int ReadByte( UWORD *x );
extern	int FlushOutput( void );
extern	int	open_output_file( void );
extern	int FillBitBuffer( void );


// *********************************************************************
//
//	STRUCTS
//
// *********************************************************************

struct local_file_header {                 /* LOCAL */
	char	ID[4];
	UBYTE	version_needed_to_extract[2];
    UWORD	general_purpose_bit_flag;
    UWORD	compression_method;
    UWORD	last_mod_file_time;
    UWORD	last_mod_file_date;
    ULONG	crc32;
	ULONG	compressed_size;
    ULONG	uncompressed_size;
    UWORD	filename_length;
    UWORD	extra_field_length;
};

struct central_directory_file_header {     /* CENTRAL */
	char	ID[4];
	UBYTE	version_made_by[2];
	UBYTE	version_needed_to_extract[2];
    UWORD	general_purpose_bit_flag;
    UWORD	compression_method;
    UWORD	last_mod_file_time;
    UWORD	last_mod_file_date;
    ULONG	crc32;
    ULONG	compressed_size;
    ULONG	uncompressed_size;
    UWORD	filename_length;
    UWORD	extra_field_length;
    UWORD	file_comment_length;
    UWORD	disk_number_start;
    UWORD	internal_file_attributes;
    ULONG	external_file_attributes;
    ULONG	relative_offset_local_header;
};

struct end_central_dir_record {            /* END CENTRAL */
	char	ID[4];
	UWORD	number_this_disk;
    UWORD	num_disk_with_start_central_dir;
    UWORD	num_entries_centrl_dir_ths_disk;
    UWORD	total_entries_central_dir;
    ULONG	size_central_directory;
    ULONG	offset_start_central_directory;
    UWORD	zipfile_comment_length;
};

union work {

	struct {
    	short Prefix_of[HSIZE + 2];      /* (8194 * sizeof(short)) */
		BYTE Suffix_of[HSIZE + 2];       /* also s-f length_nodes (smaller) */
	   	BYTE Stack[HSIZE + 2];           /* also s-f distance_nodes (smaller) */
	} shrink;

	BYTE Slide[WSIZE];
};

// *********************************************************************
//
//	MACROS
//
// *********************************************************************

#ifndef MAX
#  define MAX(a,b)   ((a) > (b) ? (a) : (b))
#endif

#ifndef MIN
#  define MIN(a,b)   ((a) < (b) ? (a) : (b))
#endif

#define prefix_of 	area->shrink.Prefix_of
#define suffix_of 	area->shrink.Suffix_of
#define stack		area->shrink.Stack
#define slide		area->Slide


#define OUTB(intc) \
{ \
	*outptr++=(BYTE)(intc); \
	if (++outcnt==OUTBUFSIZ) \
		FlushOutput(); \
}

/*
 *  macro OUTB(intc)
 *  {
 *      *outptr++ = (byte)(intc);
 *      if (++outcnt == OUTBUFSIZ)
 *          FlushOutput();
 *  }
 *
 */


#define READBIT(nbits,zdest) { \
	if( nbits>bits_left ) \
		FillBitBuffer(); \
	zdest = (int)((UWORD)bitbuf & mask_bits[nbits]); \
	bitbuf>>=nbits; \
	bits_left-=nbits; \
}

/*
 * macro READBIT(nbits,zdest)
 *  {
 *      if (nbits > bits_left)
 *          FillBitBuffer();
 *      zdest = (int)((UWORD)bitbuf & mask_bits[nbits]);
 *      bitbuf >>= nbits;
 *      bits_left -= nbits;
 *  }
 *
 */


#define PEEKBIT(nbits) (nbits>bits_left? (FillBitBuffer(),\
  (UWORD)bitbuf & mask_bits[nbits]) : \
  (UWORD)bitbuf & mask_bits[nbits])



// *********************************************************************
//
//	GLOBALS
//
// *********************************************************************

   extern int       aflag;
/* extern int       bflag;   reserved */
   extern int       cflag;
   extern int       fflag;
   extern int       jflag;
   extern int       overwrite_none;
   extern int       overwrite_all;
   extern int       force_flag;
   extern int       quietflg;
#ifdef DOS_OS2
   extern int       sflag;
#endif
   extern int       tflag;
   extern int       uflag;
   extern int       V_flag;
   extern int       process_all_files;
   extern long		csize;
   extern long		ucsize;
   extern char      *fnames[];
   extern char      **fnv;
   extern char      sig[];
   extern char      answerbuf[];
//   extern min_info  *pInfo;
   extern char      *key;
   extern ULONG     keys[];


   extern ULONG     crc32val;
   extern UWORD     mask_bits[];

   extern BYTE      *inbuf;
   extern BYTE      *inptr;
   extern int       incnt;
   extern ULONG     bitbuf;
   extern int       bits_left;
   extern BOOL		zipeof;
   extern int       zipfd;
   extern char      zipfn[];
   extern long   extra_BYTEs;
   extern long   cur_zipfile_bufstart;
   extern BYTE      *extra_field;
   extern char      local_hdr_sig[];
   extern char      central_hdr_sig[];
   extern char      end_central_sig[];
//   extern struct stat     statbuf;

   extern BYTE      *outbuf;
   extern BYTE      *outptr;
   extern BYTE      *outout;
   extern char      filename[];

   extern long   outpos;
   extern int       outcnt;
   extern int       outfd;
   extern int       mem_mode;

   extern char      *EndSigMsg;
   extern char      *CentSigMsg;
   extern char      *SeekMsg;
   extern char      *ReportMsg;


// -----------------------------------------------
//	GLOBAL STRUCTS
// -----------------------------------------------

extern	union	work							FAR *area;
extern 	struct	local_file_header				lrec;
extern 	struct	central_directory_file_header	crec;
extern 	struct	end_central_dir_record			ecrec;

extern	HGLOBAL		hExtract_files;
extern	int			extract_file_count;
extern	BOOL		extract_all;

extern	BOOL		disk_full;
extern	int			zip_in_file;			// Global handle for ZIP archive
extern	int			zip_out_file;			// Global handle for current output file

extern	char		zip_out_filename[ PATH_LEN ];

extern	char		err_msg[ 256 ];
extern	char		destination_directory[ PATH_LEN ];

extern	HWND		cwnd;

extern	BOOL	all_flag;
extern	BOOL	rep_flag;
extern	BOOL	date_flag;
extern	BOOL	dir_flag;
