// *********************************************************************
//
//	unzip.h
//
// *********************************************************************

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dir.h>
#include <direct.h>
#include <fcntl.h>
#include <io.h>
#include <sys\stat.h>

#include "os2type.h"

#define WM_COMMENCE (WM_USER+32)

#define UBYTE unsigned char
#define UWORD unsigned int
#define ULONG unsigned long

#define MAX_BITS      13                 /* used in unShrink() */
#define HSIZE         (1 << MAX_BITS)    /* size of global work area */
#define WSIZE 32768

// -----------------------------------------------
//	SIZES
// -----------------------------------------------
#define PATH_LEN		192
#define FILENAME_LEN	16

#define INBUFSIZ		16000		/* works for MS-DOS small model */
#define OUTBUFSIZ		INBUFSIZ


// -----------------------------------------------
//	COMPRESSION METHODS
// -----------------------------------------------
#define STORED            0
#define SHRUNK            1
#define REDUCED1          2
#define REDUCED2          3
#define REDUCED3          4
#define REDUCED4          5
#define IMPLODED          6
#define TOKENIZED         7
#define DEFLATED          8

// From UNREDUCE.CPP
#define DLE    144
typedef BYTE f_array[64];       /* for followers[256][64] */

// *********************************************************************
//
//	STRUCTS
//
// *********************************************************************

// -----------------------------------------------
//	ex_file struct
// -----------------------------------------------
struct   ex_file
   {
   char  file_name[ PATH_LEN ];
	char	file_node[ FILENAME_LEN ];
   };

struct local_file_header
   {                 /* LOCAL */
	char	ID[4];
	UBYTE	version_needed_to_extract[2];
   UWORD general_purpose_bit_flag;
   UWORD compression_method;
   UWORD last_mod_file_time;
   UWORD last_mod_file_date;
   ULONG crc32;
	ULONG	compressed_size;
   ULONG uncompressed_size;
   UWORD filename_length;
   UWORD extra_field_length;
   };

struct central_directory_file_header
   {     /* CENTRAL */
	char	ID[4];
	UBYTE	version_made_by[2];
	UBYTE	version_needed_to_extract[2];
   UWORD general_purpose_bit_flag;
   UWORD compression_method;
   UWORD last_mod_file_time;
   UWORD last_mod_file_date;
   ULONG crc32;
   ULONG compressed_size;
   ULONG uncompressed_size;
   UWORD filename_length;
   UWORD extra_field_length;
   UWORD file_comment_length;
   UWORD disk_number_start;
   UWORD internal_file_attributes;
   ULONG external_file_attributes;
   ULONG relative_offset_local_header;
   };

struct end_central_dir_record
   {            /* END CENTRAL */
	char	ID[4];
	UWORD	number_this_disk;
   UWORD num_disk_with_start_central_dir;
   UWORD num_entries_centrl_dir_ths_disk;
   UWORD total_entries_central_dir;
   ULONG size_central_directory;
   ULONG offset_start_central_directory;
   UWORD zipfile_comment_length;
   };

union work
   {
   struct
      {
    	short Prefix_of[HSIZE + 2];      /* (8194 * sizeof(short)) */
      BYTE  Suffix_of[HSIZE + 2];       /* also s-f length_nodes (smaller) */
      BYTE  Stack[HSIZE + 2];           /* also s-f distance_nodes (smaller) */
      } shrink;

   BYTE Slide[WSIZE];
   };

struct huft
   {
   BYTE e;               /* number of extra bits or operation */
   BYTE b;               /* number of bits in this code or subcode */
   union
      {
      UWORD n;            /* literal, length base, or distance base */
      struct huft *t;     /* pointer to next level of table */
      } v;
   };

// *********************************************************************
//
//	MACROS
//
// *********************************************************************

#ifndef MAX
#  define MAX(a,b)   ((a) > (b) ? (a) : (b))
#endif

#ifndef MIN
#  define MIN(a,b)   ((a) < (b) ? (a) : (b))
#endif

#define prefix_of 	area->shrink.Prefix_of
#define suffix_of 	area->shrink.Suffix_of
#define stack		area->shrink.Stack
#define slide		area->Slide


#define OUTB(intc) \
{ \
	*outptr++=(BYTE)(intc); \
	if (++outcnt==OUTBUFSIZ) \
		FlushOutput(); \
}

/*
 *  macro OUTB(intc)
 *  {
 *      *outptr++ = (byte)(intc);
 *      if (++outcnt == OUTBUFSIZ)
 *          FlushOutput();
 *  }
 *
 */


#define READBIT(nbits,zdest) { \
	if( nbits>bits_left ) \
		FillBitBuffer(); \
	zdest = (int)((UWORD)bitbuf & mask_bits[nbits]); \
	bitbuf>>=nbits; \
	bits_left-=nbits; \
}

/*
 * macro READBIT(nbits,zdest)
 *  {
 *      if (nbits > bits_left)
 *          FillBitBuffer();
 *      zdest = (int)((UWORD)bitbuf & mask_bits[nbits]);
 *      bitbuf >>= nbits;
 *      bits_left -= nbits;
 *  }
 *
 */

#define PEEKBIT(nbits) (nbits>bits_left? (FillBitBuffer(),\
  (UWORD)bitbuf & mask_bits[nbits]) : \
  (UWORD)bitbuf & mask_bits[nbits])

// From EXPLODE.CPP
#define NEXTBYTE    (ReadByte(&bytebuf), bytebuf)
#define NEEDBITS(n) {while(k<(n)){b|=((ULONG)NEXTBYTE)<<k;k+=8;}}
#define DUMPBITS(n) {b>>=(n);k-=(n);}

// From INFLATE.CPP
/* If BMAX needs to be larger than 16, then h and x[] should be ULONG. */
#define BMAX 16         /* maximum bit length of any code (16 for explode) */
#define N_MAX 288       /* maximum number of codes in any set */

// From UNSHRINK.CPP
/*      MAX_BITS        13   (in unzip.h; defines size of global work area)  */
#define INIT_BITS       9
#define FIRST_ENT       257
#define CLEAR           256
#define GetCode(dest)   READBIT(codesize,dest)

class TZipFile
{
public: // methods
   TZipFile(PSZ pszFileName);
   ~TZipFile()
      {}
   SHORT UnZip(PSZ pszDestinationDir = NULL);

   // *********************************************************************
   //
   // PROTOS
   //
   // *********************************************************************
   void out_msg( char *message );


   int unzip_file(PSZ pszZipFile, PSZ pszDestinationDir = NULL);
   void unShrink( void );
   void unReduce( void );
   int explode( void );
   void inflate( void );

   // -----------------------------------------------
   // FILE IO PROTOS
   // -----------------------------------------------
   int ReadByte( UWORD *x );
   int FlushOutput( void );
   int open_output_file( void );
   int FillBitBuffer( void );

   // From UNZIP.CPP
   BOOL read_zip_files( void );
   BOOL extract_this_file( char *fn );
   void unzip( void );
   void do_unzip( void );
   // int unzip_file(PSZ pszZipFile, PSZ pszDestinationDir = NULL);

   // From EXPLODE.CPP
   int get_tree( unsigned *, unsigned );
   int explode_lit8( struct huft *, struct huft *, struct huft *, int, int, int);
   int explode_lit4( struct huft *, struct huft *, struct huft *, int, int, int);
   int explode_nolit8( struct huft *, struct huft *, int, int);
   int explode_nolit4( struct huft *, struct huft *, int, int);
   //int explode( void );

   // From INFLATE.CPP
   /* Function prototypes */
   int huft_build( unsigned *, unsigned, unsigned, UWORD *, UWORD *, struct huft **, int *);
   int huft_free(struct huft *);
   void flush(unsigned);
   int inflate_codes(struct huft *, struct huft *, int, int);
   int inflate_stored(void);
   int inflate_fixed(void);
   int inflate_dynamic(void);
   int inflate_block(int *);
   int inflate_entry(void);

   // From UNREDUCE.CPP
   void LoadFollowers( void );
   void xflush( unsigned );      /* routine from inflate.c */

   // From UNSHRINK.CPP
   void  partial_clear( void );   /* local prototype */

private: // data
// *********************************************************************
//
//	GLOBALS
//
// *********************************************************************

   int       aflag;
/* int       bflag;   reserved */
   int       cflag;
   int       fflag;
   int       jflag;
   int       overwrite_none;
   int       overwrite_all;
   int       force_flag;
   int       quietflg;
#ifdef DOS_OS2
   int       sflag;
#endif
   int       tflag;
   int       uflag;
   int       V_flag;
   int       process_all_files;
   long    csize;
   long    ucsize;
   char      *fnames[];
   char      **fnv;
   char      sig[];
   char      answerbuf[];
// min_info  *pInfo;
   char      *key;
   ULONG     keys[];


   ULONG     crc32val;
   static UWORD     mask_bits[];

   BYTE      *inbuf;
   BYTE      *inptr;
   int       incnt;
   ULONG     bitbuf;
   int       bits_left;
   BOOL    zipeof;
   int       zipfd;
   char      zipfn[];
   long   extra_BYTEs;
   long   cur_zipfile_bufstart;
   BYTE      *extra_field;
   char      local_hdr_sig[];
   char      central_hdr_sig[];
   char      end_central_sig[];
// struct stat     statbuf;

   BYTE      *outbuf;
   BYTE      *outptr;
   BYTE      *outout;
   char      filename[];

   long   outpos;
   int       outcnt;
   int       outfd;
   int       mem_mode;

   char      *EndSigMsg;
   char      *CentSigMsg;
   char      *SeekMsg;
   char      *ReportMsg;

   // From UNZIP.CPP
   HGLOBAL     work_area;
// union       work                    FAR *area;
//
// struct      local_file_header          lrec;
// struct      central_directory_file_header crec;
// struct      end_central_dir_record        ecrec;

   BOOL  all_flag;
   BOOL  rep_flag;
   BOOL  date_flag;
   BOOL  dir_flag;
   BOOL  ask_path_flag;


   // -----------------------------------------------
   // FILE stuff
   // -----------------------------------------------

   BOOL     overwrite_files;

	BOOL     global_error;
	char     global_filename[ PATH_LEN ];

   // -----------------------------------------------
   // INPUT VARIABLES
   // -----------------------------------------------
	int         zip_in_file;         // Global handle for ZIP archive


	// -----------------------------------------------
   // OUTPUT VARIABLES
   // -----------------------------------------------
	int         zip_out_file;        // Global handle for current output file
   char     zip_out_filename[ PATH_LEN ];


	BOOL     disk_full;

   // -----------------------------------------------
   // FILE NAMES
   // -----------------------------------------------
   char     source_directory[ PATH_LEN ];
   char     destination_directory[ PATH_LEN ];

   char     command_line[ PATH_LEN ];
   char     zip_filename[ PATH_LEN ];
	char     option1[ 10 ];
   char     option2[ 10 ];
   char     option3[ 10 ];
   char     option4[ 10 ];

   HGLOBAL     hExtract_files;
   int         extract_file_count;
   int         total_file_count;
   BOOL        extract_all;

   // *********************************************************************
   //
   // VARS  (not related to unzipping - GUI stuff here)
   //
   // *********************************************************************
   HANDLE      hInst;
   HWND     cwnd;
   HWND     list_box;
   HWND     message_box;
   int         wwide, whigh;
   int         font_width, font_height;
   BOOL     commence;

   char     err_msg[ 256 ];

   // From INFLATE.CPP
   unsigned wp;            /* current position in slide */
   ULONG bb;                       /* bit buffer */
   unsigned bk;                    /* bits in bit buffer */
   UWORD bytebuf;
   unsigned hufts;         /* track memory usage */

   // From FILE.CPP

   // From UNSHRINK.CPP
   int   codesize;
   int   maxcode;
   int   maxcodemax;
   int   free_ent;

// -----------------------------------------------
//	GLOBAL STRUCTS
// -----------------------------------------------

   union work                    FAR *area;
   struct   local_file_header          lrec;
   struct   central_directory_file_header crec;
   struct   end_central_dir_record        ecrec;

	// From EXPLODE.CPP
   /* Tables for length and distance */
   static UWORD cplen2[];
   static UWORD cplen3[];
   static UWORD extra[];
   static UWORD cpdist4[];
   static UWORD cpdist8[];

   // From INFLATE.CPP
   /* Tables for deflate from PKZIP's appnote.txt. */
   static unsigned border[];
   static UWORD cplens[];
   static UWORD cplext[];
   static UWORD cpdist[];
   static UWORD cpdext[];

   // From UNREDUCE.CPP
   /*******************************/
   /*  UnReduce Global Variables  */
   /*******************************/

   f_array        FAR *followers;

   BYTE Slen[256];
   int factor;

   static int L_table[];
   static int D_shift[];
   static int D_mask[];
   static int B_table[];
};
